/* --------------------------------------------------------- */
/*	BASIC-E for Windows										 */
/*	Copyright(C)2006 t.hara									 */
/* ========================================================= */
/*	main window and entry point.							 */
/* --------------------------------------------------------- */

#include <tchar.h>
#include <windows.h>
#include <new>
#include <process.h>
#include "../basic_e.h"
#include "zaurus_frame.h"
#include "resource.h"

/* --------------------------------------------------------- */
namespace {

	const int WINDOW_WIDTH	= 640;
	const int WINDOW_HEIGHT	= 480;
	const int WM_MY_UPDATE	= WM_USER + 0;
	const char *p_cmd_line;

	typedef struct {
		BITMAPINFOHEADER h;
		DWORD bit_field[ 3 ];
	} c_bitmap_info;
	
	c_bitmap_info bi;
	c_zaurus_vram vram;
	c_basic_e zb( WINDOW_HEIGHT, WINDOW_WIDTH, WINDOW_HEIGHT * sizeof(unsigned short) );

	/* --------------------------------------------------------- */
	void update_client_area( HDC hdc ) {
		zb.draw( vram.get_virtual_vram() );
		vram.rotate_transfer();

		StretchDIBits( hdc,
				0, 0, WINDOW_WIDTH, WINDOW_HEIGHT,
				0, 0, WINDOW_WIDTH, WINDOW_HEIGHT,
				vram.get_physical_vram(),
				reinterpret_cast<BITMAPINFO*>( &bi ),
				DIB_RGB_COLORS, SRCCOPY );
	}

	/* --------------------------------------------------------- */
	void update_text_area( HDC hdc ) {
		zb.update_draw( vram.get_virtual_vram() );
		vram.rotate_transfer();

		StretchDIBits( hdc,
				0, 0, WINDOW_WIDTH, WINDOW_HEIGHT,
				0, 0, WINDOW_WIDTH, WINDOW_HEIGHT,
				vram.get_physical_vram(),
				reinterpret_cast<BITMAPINFO*>( &bi ),
				DIB_RGB_COLORS, SRCCOPY );
	}

	/* --------------------------------------------------------- */
	class c_callback: public c_basic_e_callback {
	private:
		CRITICAL_SECTION	cs;

	public:
		HWND h_window;

		c_callback() {
			InitializeCriticalSection( &cs );
		}

		~c_callback() {
			DeleteCriticalSection( &cs );
		}

		void sleep( unsigned int msec ) {
			Sleep( msec );
		}

		void lock( void ) {
			EnterCriticalSection( &cs );
		}

		void unlock( void ) {
			LeaveCriticalSection( &cs );
		}

		void update_screen( void ) {
			HDC hdc;
			hdc = GetDC( h_window );
			update_client_area( hdc );
			ReleaseDC( h_window, hdc );
		}
		void update_text( void ) {
			HDC hdc;
			hdc = GetDC( h_window );
			update_text_area( hdc );
			ReleaseDC( h_window, hdc );
		}
	};

	c_callback callback_instance;
	HANDLE h_thread;
	bool f_press_ctrl = false;

	/* --------------------------------------------------------- */
	unsigned int CALLBACK basic_e_thread( void* ) {
		zb.run( p_cmd_line );
		PostMessage( callback_instance.h_window, WM_CLOSE, 0, 0 );
		return 0;
	}

	/* --------------------------------------------------------- */
	LRESULT CALLBACK window_procedure( HWND h_window, UINT message, WPARAM w_parameter, LPARAM l_parameter ) {
		static BYTE key_state[ 256 ];
		WORD key_buffer = 0;

		switch( message ) {
		case WM_KEYDOWN:
			if( w_parameter == VK_CONTROL ) {
				key_state[ w_parameter & 0xFF ] = 0x80;
				f_press_ctrl = true;
				return 0;
			}
			if( w_parameter == VK_INSERT ) {
				zb.key( c_basic_e::CC_INS );
				return 0;
			}
			if( w_parameter == VK_CAPITAL ) {
				key_state[ VK_CAPITAL ] = key_state[ VK_CAPITAL ] ^ 1;
			}
			else {
				key_state[ w_parameter & 0xFF ] = 0x80;
				if( w_parameter >= VK_LEFT && w_parameter <= VK_DOWN ) {
					/* It is cursor key. */
					w_parameter = (w_parameter & 0x0F) | 0x1C;
					if( (w_parameter == 0x1C) || (w_parameter == 0x1F) ) {
						w_parameter = w_parameter ^ 0x3;
					}
					zb.key( (int)w_parameter );
					return 0;
				}
				if( w_parameter == VK_DELETE ) {
					zb.key( c_basic_e::CC_DEL );
					return 0;
				}
			}
			if( f_press_ctrl && (w_parameter == 'C') ) {
				f_press_ctrl = false;
				zb.force_break();
			}
			else if( ToAscii( (UINT)w_parameter, (UINT)((l_parameter >> 16) & 0xFF), key_state, &key_buffer, 0 ) == 1 ) {
				zb.key( key_buffer );
			}
			return 0;
		case WM_KEYUP:
			key_state[ w_parameter & 0xFF ] = key_state[ w_parameter & 0xFF ] ^ 0x80;
			f_press_ctrl = false;
			return 0;
		case WM_PAINT:
			PAINTSTRUCT paint_struct;
			BeginPaint( h_window, &paint_struct );
			update_client_area( paint_struct.hdc );
			EndPaint( h_window, &paint_struct );
			return 0;
		case WM_MY_UPDATE:
			InvalidateRect( h_window, NULL, FALSE );
			return 0;
		case WM_LBUTTONDOWN:
		case WM_CLOSE:
			zb.terminate();
			WaitForSingleObject( h_thread, 3000 );
			CloseHandle( h_thread );
			DestroyWindow( h_window );
			return 0;
		case WM_DESTROY:
			PostQuitMessage( 0 );
			return 0;
		}
		return DefWindowProc( h_window, message, w_parameter, l_parameter );
	}

	/* --------------------------------------------------------- */
	void initialize_main_window( int show_flag ) {

		bi.h.biBitCount			= 16;
		bi.h.biClrImportant		= 0;
		bi.h.biClrUsed			= 0;
		bi.h.biCompression		= BI_BITFIELDS;
		bi.h.biHeight			= WINDOW_HEIGHT;
		bi.h.biPlanes			= 1;
		bi.h.biSize				= sizeof(bi) + sizeof(DWORD) * 3;
		bi.h.biSizeImage		= WINDOW_WIDTH * WINDOW_HEIGHT * 2;
		bi.h.biWidth			= WINDOW_WIDTH;
		bi.h.biXPelsPerMeter	= 100;
		bi.h.biYPelsPerMeter	= 100;
		bi.bit_field[ 0 ]		= 0xF800;	/* red		*/
		bi.bit_field[ 1 ]		= 0x07E0;	/* green	*/
		bi.bit_field[ 2 ]		= 0x001F;	/* blue		*/

		zb.initialize( &callback_instance );
		vram.initialize();

		WNDCLASSEX	wc;
		wc.cbClsExtra		= 0;
		wc.cbSize			= sizeof(wc);
		wc.cbWndExtra		= 0;
		wc.hbrBackground	= NULL;
		wc.hCursor			= LoadCursor( NULL, IDC_ARROW );
		wc.hIcon			= LoadCursor( GetModuleHandle( NULL ), MAKEINTRESOURCE(IDI_MAIN) );
		wc.hIconSm			= LoadCursor( GetModuleHandle( NULL ), MAKEINTRESOURCE(IDI_MAIN) );
		wc.hInstance		= GetModuleHandle( NULL );
		wc.lpfnWndProc		= window_procedure;
		wc.lpszClassName	= TEXT( "BASIC-E window class" );
		wc.lpszMenuName		= NULL;
		wc.style			= 0;
		if( ! RegisterClassEx( &wc ) ) {
			throw( c_basic_e_exception( c_basic_e_exception::SYSTEM_ERROR ) );
		}

		RECT r = { 0, 0, 640, 480 };
		AdjustWindowRect( &r, WS_POPUP | WS_CAPTION | WS_SYSMENU, FALSE );
		callback_instance.h_window = CreateWindow( wc.lpszClassName, TEXT( "BASIC-E for Windows" ),
				WS_POPUP | WS_CAPTION | WS_SYSMENU, 
				CW_USEDEFAULT, CW_USEDEFAULT, r.right - r.left, r.bottom - r.top, NULL, NULL, wc.hInstance, NULL );
		if( callback_instance.h_window == NULL ) {
			throw( c_basic_e_exception( c_basic_e_exception::SYSTEM_ERROR ) );
		}

		ShowWindow( callback_instance.h_window, show_flag );
		UpdateWindow( callback_instance.h_window );

		unsigned int thread_id;
		h_thread = (HANDLE) _beginthreadex( NULL, 0, basic_e_thread, NULL, 0, &thread_id );
	}

}

/* --------------------------------------------------------- */
INT WINAPI _tWinMain( HINSTANCE , HINSTANCE , LPTSTR p_cmd, int show_flag ) {
	MSG message;
	BOOL result;

	p_cmd_line = p_cmd;
	try {
		initialize_main_window( show_flag );

		for( ;; ) {
			result = GetMessage( &message, NULL, 0, 0 );
			if( result == -1 ) {
				throw( c_basic_e_exception( c_basic_e_exception::SYSTEM_ERROR ) );
			}
			if( result == 0 ) {
				break;
			}
			TranslateMessage( &message );
			DispatchMessage( &message );
		}
	}
	catch( c_basic_e_exception &e ) {
		MessageBox( NULL, e.what(), TEXT( "ERROR" ), MB_OK | MB_ICONWARNING );
	}
	catch( std::bad_alloc& ) {
		MessageBox( NULL, TEXT( "Not enough memory." ), TEXT( "ERROR" ), MB_OK | MB_ICONWARNING );
	}
	
	return 0;
}
