/* --------------------------------------------------------- */
/*	BASIC-E													 */
/*	Copyright(C)2006 t.hara									 */
/* ========================================================= */
/*	BASIC-E body											 */
/* --------------------------------------------------------- */

#include "basic_e.h"
#include "zb_error.h"
#include <cstdio>
#include "debug.h"

/* --------------------------------------------------------- */
void c_basic_e::initialize( c_basic_e_callback *p ) {

	resource.p_graphic_screen	= &graphic_screen;
	resource.p_text_screen		= &text_screen;
	resource.p_text_editor		= &text_editor;
	resource.p_interpreter		= &interpreter;
	resource.p_variable_manager	= &variable_manager;
	resource.p_keyboard			= &keyboard;
	resource.error_code			= c_zb_error::NO_PROBLEM;
	resource.p_callback			= p;

	interpreter.set_system_resource( &resource );
	text_editor.set_system_resource( &resource );
	keyboard.set_system_resource( &resource );

	/* create and initialize screen memories. */
	graphic_screen.create( width, height );
	text_screen.create( width, height, line_step );
	text_screen.set_graphic( &graphic_screen );

	text_editor.set_edit_mode( false );
	text_editor.put_str( "BASIC-E v0.0.6\xD\xA" );
	text_editor.put_str( "Copyright(C)2006 t.hara\xD\xA" );
	text_editor.put_str( "Ok\xD\xA" );
	text_editor.set_edit_mode( true );
}

/* --------------------------------------------------------- */
void c_basic_e::key( int key_code ) {

	keyboard.push_key( key_code );
}

/* --------------------------------------------------------- */
void c_basic_e::force_break( void ) {

	resource.error_code = c_zb_error::FORCE_BREAK;
}

/* --------------------------------------------------------- */
void c_basic_e::draw( unsigned short *p_vram ) {

	text_screen.draw( p_vram );
}

/* --------------------------------------------------------- */
void c_basic_e::update_draw( unsigned short *p_vram ) {

	text_screen.update( p_vram );
}

/* --------------------------------------------------------- */
void c_basic_e::run( const char *p_line ) {
	int key_code, i;
	char s_file_name[ 260 ];

	dlog.open();
	dlog.printf( "BASIC-E start\n" );
	resource.f_running = true;
	if( p_line != 0 && p_line[0] != '\0' ) {
		dlog.printf( "detect command line parameter.\n" );
		if( p_line[0] == '\"' ) {
			p_line++;
			strncpy( s_file_name, p_line, sizeof(s_file_name) );
			for( i = 0; i < static_cast<int>(sizeof(s_file_name)); i++ ) {
				if( s_file_name[i] == '\"' ) {
					s_file_name[i] = '\0';
				}
			}
		}
		else {
			strncpy( s_file_name, p_line, sizeof(s_file_name) );
		}
		dlog.printf( "command line parameter is:\n \"%s\"\n", s_file_name );
		resource.p_interpreter->list.load( s_file_name );
		if( ! resource.error_code ) {
			dlog.printf( "success program load.\n" );
			const char *p_string = "run";
			c_parser parser;
			parser.set_system_resource( &resource );
			parser.check_line_no( p_string );
			parser.precompile( p_string, 0 );
			resource.p_text_editor->enable_cursor( false );
			resource.p_text_editor->set_edit_mode( false );
			resource.p_interpreter->run( &parser );
			resource.p_text_editor->set_edit_mode( true );
			resource.p_text_editor->enable_cursor( true );
			resource.p_text_editor->show_cursor();
			resource.p_callback->update_text();
		}
		else {
			dlog.printf( "failed program load.\n" );
			resource.error_code = c_zb_error::NO_PROBLEM;
		}
	}

	while( resource.f_running ) {
		key_code = keyboard.pop_key();
		if( key_code ) {
			text_editor.key( key_code );
		}
		else {
			resource.p_callback->sleep( 10 );
		}
	}
}
