/* --------------------------------------------------------- */
/*	BASIC-E													 */
/*	Copyright(C)2006 t.hara									 */
/* ========================================================= */
/*	interpreter												 */
/* --------------------------------------------------------- */

#include "system_resource.h"
#include "parser.h"
#include "variant_value.h"
#include "interpreter.h"
#include "command.h"
#include "function.h"
#include "zb_error.h"
#include "variable_manager.h"
#include "debug.h"
#include <cstring>
#include <cstdlib>

using namespace std;

/* --------------------------------------------------------- */
int c_expression::class_change( c_value& result, int operand_type ) {
	int result_type  = result.get()->get_type();

	if( result_type < c_base_type::TYPE_REAL ) {
		if( operand_type == c_base_type::TYPE_REAL ) {
			result.change_real( result.get()->get_real() );
		}
		else if( operand_type == c_base_type::TYPE_STRING ) {
			result.convert_to_string();
		}
		else if( operand_type == c_base_type::TYPE_INTEGER ) {
			result.change_integer( result.get()->get_integer() );
		}
	}
	else if( result_type == c_base_type::TYPE_STRING ) {
	}
	else {
		if( operand_type == c_base_type::TYPE_STRING ) {
			return c_zb_error::ERROR_TYPE_MISMATCH;
		}
	}
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_expression::term( c_value& term_value, const unsigned char* &p_code, bool &f_continue ) {
	int error_code = c_zb_error::NO_PROBLEM;
	int n;
	double r;
	int elements[ MAX_ARRAY_DIMENTION ], num_elements;

	switch( p_code[ 0 ] ) {
	case c_parser::OP_L_BLACKET:
		p_code++;
		{
			c_expression expression( p_res );
			error_code = expression.expression( term_value, p_code );
		}
		if( error_code != c_zb_error::NO_PROBLEM ) return error_code;
		if( p_code[ 0 ] != c_parser::OP_R_BLACKET ) return c_zb_error::ERROR_NOT_FOUND_R_BLACKET;
		p_code++;
		return c_zb_error::NO_PROBLEM;
	case c_parser::OP_PLUS:		/* sign PLUS */
		p_code++;
		error_code = term( term_value, p_code, f_continue );
		if( error_code != c_zb_error::NO_PROBLEM ) return error_code;
		return term_value.get()->plus();
	case c_parser::OP_MINUS:	/* sign MINUS */
		p_code++;
		error_code = term( term_value, p_code, f_continue );
		if( error_code != c_zb_error::NO_PROBLEM ) return error_code;
		return term_value.get()->minus();
	case c_parser::OP_NOT:
		p_code++;
		term( term_value, p_code, f_continue );
		return term_value.get()->not();
	case c_parser::PC_FUNCTION:
		memcpy( &n, p_code + 1, sizeof(n) );
		p_code += 1 + sizeof(n);
		if( p_code[0] != c_parser::OP_L_BLACKET ) return c_zb_error::ERROR_SYNTAX;
		p_code++;
		(function.get_entry(n))( term_value, p_code, p_res );
		if( p_code[0] != c_parser::OP_R_BLACKET ) return c_zb_error::ERROR_ILLEGAL_PARAMETER;
		p_code++;
		f_continue = ( p_res->error_code == c_zb_error::NO_PROBLEM );
		return p_res->error_code;
	case c_parser::PC_INUMBER:
		p_code++;
		memcpy( &n, p_code, sizeof(int) );
		p_code += sizeof(int);
		term_value.change_integer( n );
		return c_zb_error::NO_PROBLEM;
	case c_parser::PC_RNUMBER:
		p_code++;
		memcpy( &r, p_code, sizeof(double) );
		p_code += sizeof(double);
		term_value.change_real( r );
		return c_zb_error::NO_PROBLEM;
	case c_parser::PC_STRING:
		n = p_code[1];
		term_value.change_string( (char*)p_code + 2, n );
		p_code += n + 2;
		return c_zb_error::NO_PROBLEM;
	case c_parser::PC_IVAR:
		{
			p_code++;
			c_variable_integer *p;
			p = reinterpret_cast<c_variable_integer*>( c_parser::get_ptr( p_code ) );
			p_code += sizeof(void*);
			term_value.change_integer( p->get() );
		}
		return c_zb_error::NO_PROBLEM;
	case c_parser::PC_RVAR:
		{
			p_code++;
			c_variable_real *p;
			p = reinterpret_cast<c_variable_real*>( c_parser::get_ptr( p_code ) );
			p_code += sizeof(void*);
			term_value.change_real( p->get() );
		}
		return c_zb_error::NO_PROBLEM;
	case c_parser::PC_SVAR:
		{
			p_code++;
			c_variable_string *p;
			p = reinterpret_cast<c_variable_string*>( c_parser::get_ptr( p_code ) );
			p_code += sizeof(void*);
			term_value.change_string( p->get(), p->get_length() );
		}
		return c_zb_error::NO_PROBLEM;
	case c_parser::PC_IARRAY:
		{
			p_code++;
			c_variable_integer_array *p;
			int v;
			p = reinterpret_cast<c_variable_integer_array*>( c_parser::get_ptr( p_code ) );
			p_code += sizeof(void*);
			num_elements = p_res->p_interpreter->dimension_check( p_code, elements ); 
			if( num_elements == 0 ) return p_res->error_code;
			p_res->error_code = p->get( num_elements, elements, v );
			term_value.change_integer( v );
		}
		return c_zb_error::NO_PROBLEM;
	case c_parser::PC_RARRAY:
		{
			p_code++;
			c_variable_real_array *p;
			double v;
			p = reinterpret_cast<c_variable_real_array*>( c_parser::get_ptr( p_code ) );
			p_code += sizeof(void*);
			num_elements = p_res->p_interpreter->dimension_check( p_code, elements ); 
			if( num_elements == 0 ) return p_res->error_code;
			p_res->error_code = p->get( num_elements, elements, v );
			term_value.change_real( v );
		}
		return c_zb_error::NO_PROBLEM;
	case c_parser::PC_SARRAY:
		{
			p_code++;
			c_variable_string_array *p;
			const char *p_string_value;
			int length;
			p = reinterpret_cast<c_variable_string_array*>( c_parser::get_ptr( p_code ) );
			p_code += sizeof(void*);
			num_elements = p_res->p_interpreter->dimension_check( p_code, elements ); 
			if( num_elements == 0 ) return p_res->error_code;
			p_res->error_code = p->get( num_elements, elements, p_string_value, length );
			term_value.change_string( p_string_value, length );
		}
		return c_zb_error::NO_PROBLEM;
	default:
		break;
	}
	return error_code;
}

/* --------------------------------------------------------- */
int c_expression::power( c_value& result, const unsigned char* &p_code, bool &f_continue ) {
	int error_code = c_zb_error::NO_PROBLEM;
	c_value term_value;

	error_code = term( result, p_code, f_continue );
	while( f_continue ) {
		switch( p_code[0] ) {
		case c_parser::OP_POWER:
			p_code++;
			/*	This operator processes it from the right side.	*/
			error_code = power( term_value, p_code, f_continue );
			if( error_code != c_zb_error::NO_PROBLEM ) return error_code;
			error_code = class_change( result, term_value.get()->get_type() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			error_code = result.get()->power( term_value.get() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			break;
		default:
			return error_code;
		}
	}
	return error_code;
}

/* --------------------------------------------------------- */
int c_expression::multiplication( c_value& result, const unsigned char* &p_code, bool &f_continue ) {
	int error_code = c_zb_error::NO_PROBLEM;
	c_value term_value;

	error_code = power( result, p_code, f_continue );
	while( f_continue ) {
		switch( p_code[0] ) {
		case c_parser::OP_MUL:
			p_code++;
			error_code = power( term_value, p_code, f_continue );
			if( error_code != c_zb_error::NO_PROBLEM ) return error_code;
			error_code = class_change( result, term_value.get()->get_type() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			error_code = result.get()->multiplication( term_value.get() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			break;
		case c_parser::OP_DIV:
			p_code++;
			error_code = power( term_value, p_code, f_continue );
			if( error_code != c_zb_error::NO_PROBLEM ) return error_code;
			error_code = class_change( result, term_value.get()->get_type() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			error_code = result.get()->division( term_value.get() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			break;
		case c_parser::OP_MOD:
			p_code++;
			error_code = power( term_value, p_code, f_continue );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				return error_code;
			}
			error_code = class_change( result, term_value.get()->get_type() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			error_code = result.get()->remainder( term_value.get() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			break;
		default:
			return error_code;
		}
	}
	return error_code;
}

/* --------------------------------------------------------- */
int c_expression::shift( c_value& result, const unsigned char* &p_code, bool &f_continue ) {
	int error_code = c_zb_error::NO_PROBLEM;
	c_value term_value;

	error_code = multiplication( result, p_code, f_continue );
	while( f_continue ) {
		switch( p_code[0] ) {
		case c_parser::OP_LSHIFT:
			p_code++;
			error_code = multiplication( term_value, p_code, f_continue );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				return error_code;
			}
			error_code = class_change( result, term_value.get()->get_type() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			error_code = result.get()->lshift( term_value.get() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			break;
		case c_parser::OP_RSHIFT:
			p_code++;
			error_code = multiplication( term_value, p_code, f_continue );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				return error_code;
			}
			error_code = class_change( result, term_value.get()->get_type() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			error_code = result.get()->rshift( term_value.get() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			break;
		default:
			return error_code;
		}
	}
	return error_code;
}

/* --------------------------------------------------------- */
int c_expression::addition( c_value& result, const unsigned char* &p_code, bool &f_continue ) {
	int error_code = c_zb_error::NO_PROBLEM;
	c_value term_value;

	error_code = shift( result, p_code, f_continue );
	while( f_continue ) {
		switch( p_code[0] ) {
		case c_parser::OP_PLUS:
			p_code++;
			error_code = shift( term_value, p_code, f_continue );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				return error_code;
			}
			error_code = class_change( result, term_value.get()->get_type() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			error_code = result.get()->addition( term_value.get() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			break;
		case c_parser::OP_MINUS:
			p_code++;
			error_code = shift( term_value, p_code, f_continue );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				return error_code;
			}
			error_code = class_change( result, term_value.get()->get_type() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			error_code = result.get()->subtraction( term_value.get() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			break;
		default:
			return error_code;
		}
	}
	return error_code;
}

/* --------------------------------------------------------- */
int c_expression::comparator( c_value& result, const unsigned char* &p_code, bool &f_continue ) {
	int n = 0;
	int error_code = c_zb_error::NO_PROBLEM;
	c_value term_value;

	error_code = addition( result, p_code, f_continue );
	while( f_continue ) {
		switch( p_code[0] ) {
		case c_parser::OP_EQU:
			p_code++;
			error_code = addition( term_value, p_code, f_continue );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				return error_code;
			}
			error_code = class_change( result, term_value.get()->get_type() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			error_code = result.get()->equ( n, term_value.get() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			break;
		case c_parser::OP_NOT_EQU:
			p_code++;
			error_code = addition( term_value, p_code, f_continue );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				return error_code;
			}
			error_code = class_change( result, term_value.get()->get_type() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			error_code = result.get()->not_equ( n, term_value.get() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			break;
		case c_parser::OP_GT:
			p_code++;
			error_code = addition( term_value, p_code, f_continue );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				return error_code;
			}
			error_code = class_change( result, term_value.get()->get_type() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			error_code = result.get()->gt( n, term_value.get() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			break;
		case c_parser::OP_GT_EQU:
			p_code++;
			error_code = addition( term_value, p_code, f_continue );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				return error_code;
			}
			error_code = class_change( result, term_value.get()->get_type() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			error_code = result.get()->gt_equ( n, term_value.get() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			break;
		case c_parser::OP_LT:
			p_code++;
			error_code = addition( term_value, p_code, f_continue );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				return error_code;
			}
			error_code = class_change( result, term_value.get()->get_type() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			error_code = result.get()->lt( n, term_value.get() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			break;
		case c_parser::OP_LT_EQU:
			p_code++;
			error_code = addition( term_value, p_code, f_continue );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				return error_code;
			}
			error_code = class_change( result, term_value.get()->get_type() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			error_code = result.get()->lt_equ( n, term_value.get() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			break;
		default:
			return error_code;
		}
		result.change_integer( n );
	}
	return error_code;
}

/* --------------------------------------------------------- */
int c_expression::and( c_value& result, const unsigned char* &p_code, bool &f_continue ) {
	int error_code = c_zb_error::NO_PROBLEM;
	c_value term_value;

	error_code = comparator( result, p_code, f_continue );
	while( f_continue ) {
		switch( p_code[0] ) {
		case c_parser::OP_AND:
			p_code++;
			error_code = comparator( term_value, p_code, f_continue );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				return error_code;
			}
			error_code = class_change( result, term_value.get()->get_type() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			error_code = result.get()->and( term_value.get() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			break;
		default:
			return error_code;
		}
	}
	return error_code;
}

/* --------------------------------------------------------- */
int c_expression::or( c_value& result, const unsigned char* &p_code, bool &f_continue ) {
	int error_code = c_zb_error::NO_PROBLEM;
	c_value term_value;

	error_code = and( result, p_code, f_continue );
	while( f_continue ) {
		switch( p_code[0] ) {
		case c_parser::OP_OR:
			p_code++;
			error_code = and( term_value, p_code, f_continue );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				return error_code;
			}
			error_code = class_change( result, term_value.get()->get_type() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			error_code = result.get()->or( term_value.get() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			break;
		case c_parser::OP_XOR:
			p_code++;
			error_code = and( term_value, p_code, f_continue );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				return error_code;
			}
			error_code = class_change( result, term_value.get()->get_type() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			error_code = result.get()->xor( term_value.get() );
			if( error_code != c_zb_error::NO_PROBLEM ) {
				f_continue = false;
				return error_code;
			}
			break;
		default:
			return error_code;
		}
	}
	return error_code;
}

/* --------------------------------------------------------- */
int c_expression::expression( c_value &result, const unsigned char* &p_code ) {
	bool f_continue = true;

	return or( result, p_code, f_continue );
}

/* --------------------------------------------------------- */
void c_interpreter::cmd_goto( const unsigned char* &p_code ) {
	c_variable_label *p_label;

	if( p_code[1] != c_parser::PC_LABEL ) {
		p_res->error_code = c_zb_error::ERROR_ILLEGAL_PARAMETER;
		return;
	}
	p_code += 2;
	p_label = reinterpret_cast<c_variable_label*>( c_parser::get_ptr( p_code ) );
	if( p_label->get_line() == 0 ) {
		p_res->error_code = c_zb_error::ERROR_INVALID_LABEL;
		return;
	}
	p_code = p_label->get_line()->get_code();
	if( p_code == 0 ) {
		p_res->error_code = c_zb_error::ERROR_INVALID_LABEL;
		return;
	}
	p_code += p_label->get();
	list.goto_ptr( p_label->get_line() );
}

/* --------------------------------------------------------- */
void c_interpreter::cmd_if( const unsigned char* &p_code ) {
	c_variable_label *p_label;
	c_value result;
	c_expression expression( p_res );

	p_code++;
	expression.expression( result, p_code );	
	if( p_res->error_code != c_zb_error::NO_PROBLEM ) {
		return;
	}
	/* case of true */
	if( (p_code[0] != c_parser::PC_THEN) && (p_code[0] != c_parser::PC_GOTO) && (p_code[1] != c_parser::PC_LABEL) ) {
		p_res->error_code = c_zb_error::ERROR_SYNTAX;
		return;
	}
	p_label = reinterpret_cast<c_variable_label*>( c_parser::get_ptr( p_code + 2 ) );
	if( result.get()->get_integer() ) {
		if( p_label->get_line() == 0 ) {
			p_res->error_code = c_zb_error::ERROR_INVALID_LABEL;
			return;
		}
		p_code = p_label->get_line()->get_code();
		if( p_code == 0 ) {
			p_res->error_code = c_zb_error::ERROR_INVALID_LABEL;
			return;
		}
		p_code += p_label->get();
		list.goto_ptr( p_label->get_line() );
		return;
	}
	/* case of false */
	p_code += 2 + sizeof(void*);
	if( (p_code[0] != c_parser::PC_ELSE) || (p_code[1] != c_parser::PC_LABEL) ) {
		return;
	}
	p_label = reinterpret_cast<c_variable_label*>( c_parser::get_ptr( p_code + 2 ) );
	if( p_label->get_line() == 0 ) {
		p_res->error_code = c_zb_error::ERROR_INVALID_LABEL;
		return;
	}
	p_code = p_label->get_line()->get_code();
	if( p_code == 0 ) {
		p_res->error_code = c_zb_error::ERROR_INVALID_LABEL;
		return;
	}
	p_code += p_label->get();
	list.goto_ptr( p_label->get_line() );
}

/* --------------------------------------------------------- */
void c_interpreter::cmd_for( const unsigned char* &p_code ) {
	c_expression expression( p_res );
	c_value result;
	c_for_info info;

	p_code++;	/* skip PC_FOR */
	/* check loop variable */
	if( p_code[0] != c_parser::PC_IVAR ) {
		p_res->error_code = c_zb_error::ERROR_ILLEGAL_LOOP_COUNTER;
		return;
	}
	info.p_loop_counter = process_ivar( p_code );
	if( p_res->error_code != c_zb_error::NO_PROBLEM ) return;
	if( p_code[0] != c_parser::PC_TO ) {
		p_res->error_code = c_zb_error::ERROR_ILLEGAL_PARAMETER;
		return;
	}
	p_code++;	/* skip PC_TO */
	expression.expression( result, p_code );
	info.end_point = result.get()->get_integer();
	if( p_code[0] == c_parser::PC_STEP ) {
		p_code++;	/* skip PC_STEP */
		expression.expression( result, p_code );
		info.next_step = result.get()->get_integer();
	}
	else {
		info.next_step = 1;
	}
	info.p_code = p_code;
	info.p_line = list.get_current_line();

	if( for_info.is_full() ) {
		p_res->error_code = c_zb_error::ERROR_FOR_STACK_OVERFLOW;
		return;
	}
	for_info.push( info );
}

/* --------------------------------------------------------- */
void c_interpreter::cmd_next( const unsigned char* &p_code ) {
	c_for_info info;

	p_code++;

	if( ! for_info.is_buffered() ) {
		p_res->error_code = c_zb_error::ERROR_NEXT_WITHOUT_FOR;
		return;
	}
	info = for_info.pop();

	info.p_loop_counter->set( info.p_loop_counter->get() + info.next_step );
	if( info.next_step > 0 ) {
		/* increment counter */
		if( info.p_loop_counter->get() <= info.end_point ) {
			p_code = info.p_code;
			list.goto_ptr( info.p_line );
			for_info.push( info );
		}
	}
	else if( info.next_step < 0 ) {
		/* increment counter */
		if( info.p_loop_counter->get() >= info.end_point ) {
			p_code = info.p_code;
			list.goto_ptr( info.p_line );
			for_info.push( info );
		}
	}
	else {
		/* infinite loop */
		if( info.p_loop_counter->get() == info.end_point ) {
			p_code = info.p_code;
			list.goto_ptr( info.p_line );
			for_info.push( info );
		}
	}
}

/* --------------------------------------------------------- */
void c_interpreter::cmd_gosub( const unsigned char* &p_code ) {
	c_gosub_info info;
	c_variable_label *p_label;

	if( p_code[1] != c_parser::PC_LABEL ) {
		p_res->error_code = c_zb_error::ERROR_ILLEGAL_PARAMETER;
		return;
	}
	p_code += 2;
	p_label = reinterpret_cast<c_variable_label*>( c_parser::get_ptr( p_code ) );

	info.p_code = p_code + sizeof(void*);
	info.p_line = list.get_current_line();

	if( p_label->get_line() == 0 ) {
		p_res->error_code = c_zb_error::ERROR_INVALID_LABEL;
		return;
	}
	p_code = p_label->get_line()->get_code();
	if( p_code == 0 ) {
		p_res->error_code = c_zb_error::ERROR_INVALID_LABEL;
		return;
	}
	p_code += p_label->get();
	list.goto_ptr( p_label->get_line() );

	if( gosub_info.is_full() ) {
		p_res->error_code = c_zb_error::ERROR_GOSUB_STACK_OVERFLOW;
		return;
	}
	gosub_info.push( info );
}

/* --------------------------------------------------------- */
void c_interpreter::cmd_return( const unsigned char* &p_code ) {
	c_gosub_info info;

	if( ! gosub_info.is_buffered() ) {
		p_res->error_code = c_zb_error::ERROR_RETURN_WITHOUT_GOSUB;
		return;
	}
	info = gosub_info.pop();
	p_code = info.p_code;
	list.goto_ptr( info.p_line );
}

/* --------------------------------------------------------- */
int c_interpreter::dimension_check( const unsigned char* &p_code, int *p_elements ) {
	c_expression expression( p_res );
	c_value result;
	int elements;

	if( *p_code != c_parser::OP_L_BLACKET ) {
		p_res->error_code = c_zb_error::ERROR_SYNTAX;
		return 0;
	}
	/* skip '('	*/
	p_code++;

	for( elements = 0; elements < MAX_ARRAY_DIMENTION; ) {
		expression.expression( result, p_code );
		if( p_res->error_code != c_zb_error::NO_PROBLEM ) return 0;
		p_elements[ elements ] = result.get()->get_integer();
		elements++;
		if( *p_code != c_parser::OP_COMMA ) {
			break;
		}
		/* skip ','	*/
		p_code++;
	}
	
	if( *p_code != c_parser::OP_R_BLACKET ) {
		p_res->error_code = c_zb_error::ERROR_SYNTAX;
		return 0;
	}
	/* skip ')'	*/
	p_code++;

	return elements;
}

/* --------------------------------------------------------- */
void c_interpreter::dimension_alloc( const unsigned char* &p_code ) {
	c_variable_integer_array *p_integer_array;
	c_variable_real_array *p_real_array;
	c_variable_string_array *p_string_array;
	int i;
	int elements[ MAX_ARRAY_DIMENTION ], num_elements;

	for( i = 0; !(c_parser::is_command_end( p_code[0] )); i++ ) {
		switch( p_code[0] ) {
		case c_parser::PC_IARRAY:
			p_code++;
			p_integer_array = reinterpret_cast<c_variable_integer_array*>( c_parser::get_ptr( p_code ) );
			p_code += sizeof(void*);
			num_elements = dimension_check( p_code, elements );
			if( (num_elements == 0) || (p_res->error_code != c_zb_error::NO_PROBLEM) ) {
				break;
			}
			p_res->error_code = p_integer_array->dim( num_elements, elements );
			break;
		case c_parser::PC_RARRAY:
			p_code++;
			p_real_array = reinterpret_cast<c_variable_real_array*>( c_parser::get_ptr( p_code ) );
			p_code += sizeof(void*);
			num_elements = dimension_check( p_code, elements );
			if( (num_elements == 0) || (p_res->error_code != c_zb_error::NO_PROBLEM) ) {
				break;
			}
			p_res->error_code = p_real_array->dim( num_elements, elements );
			break;
		case c_parser::PC_SARRAY:
			p_code++;
			p_string_array = reinterpret_cast<c_variable_string_array*>( c_parser::get_ptr( p_code ) );
			p_code += sizeof(void*);
			num_elements = dimension_check( p_code, elements );
			if( (num_elements == 0) || (p_res->error_code != c_zb_error::NO_PROBLEM) ) {
				break;
			}
			p_res->error_code = p_string_array->dim( num_elements, elements );
			break;
		default:
			p_res->error_code = c_zb_error::ERROR_SYNTAX;
			return;
		}
		if( p_code[0] != c_parser::OP_COMMA ) break;
		p_code++;
	}
	if( !(c_parser::is_command_end( p_code[0] )) ) {
		p_res->error_code = c_zb_error::ERROR_SYNTAX;
	}
}

/* --------------------------------------------------------- */
void c_interpreter::dimension_free( void ) {

	p_res->p_variable_manager->array_clear();
}

/* --------------------------------------------------------- */
c_variable_integer *c_interpreter::process_ivar( const unsigned char* &p_code ) {
	c_variable_integer *p_integer;
	c_expression expression( p_res );
	c_value result;

	p_code++;
	p_integer = reinterpret_cast<c_variable_integer*>( c_parser::get_ptr( p_code ) );
	p_code += sizeof(void*);
	if( p_code[0] != c_parser::OP_EQU ) {
		p_res->error_code = c_zb_error::ERROR_SYNTAX;
		return p_integer;
	}
	p_code++;
	p_res->error_code = expression.expression( result, p_code );
	if( p_res->error_code == c_zb_error::NO_PROBLEM ) {
		p_integer->set( result.get()->get_integer() );
	}
	return p_integer;
}

/* --------------------------------------------------------- */
void c_interpreter::process_rvar( const unsigned char* &p_code ) {
	c_variable_real *p_real;
	c_expression expression( p_res );
	c_value result;

	p_code++;
	p_real = reinterpret_cast<c_variable_real*>( c_parser::get_ptr( p_code ) );
	p_code += sizeof(void*);
	if( p_code[0] != c_parser::OP_EQU ) {
		p_res->error_code = c_zb_error::ERROR_SYNTAX;
		return;
	}
	p_code++;
	p_res->error_code = expression.expression( result, p_code );
	if( p_res->error_code == c_zb_error::NO_PROBLEM ) {
		p_real->set( result.get()->get_real() );
	}
}

/* --------------------------------------------------------- */
void c_interpreter::process_svar( const unsigned char* &p_code ) {
	c_variable_string *p_string;
	c_expression expression( p_res );
	c_value result;
	const char *p_string_value;
	int length;

	p_code++;
	p_string = reinterpret_cast<c_variable_string*>( c_parser::get_ptr( p_code ) );
	p_code += sizeof(void*);
	if( p_code[0] != c_parser::OP_EQU ) {
		p_res->error_code = c_zb_error::ERROR_SYNTAX;
		return;
	}
	p_code++;
	p_res->error_code = expression.expression( result, p_code );
	if( p_res->error_code == c_zb_error::NO_PROBLEM ) {
		result.get()->get_string( p_string_value, length );
		p_string->set( p_string_value, length );
	}
}

/* --------------------------------------------------------- */
void c_interpreter::process_iarray( const unsigned char* &p_code ) {
	c_variable_integer_array *p_integer_array;
	int elements[ MAX_ARRAY_DIMENTION ], num_elements;
	c_expression expression( p_res );
	c_value result;

	p_code++;
	p_integer_array = reinterpret_cast<c_variable_integer_array*>( c_parser::get_ptr( p_code ) );
	p_code += sizeof(void*);
	num_elements = dimension_check( p_code, elements );
	if( (num_elements == 0) || (p_res->error_code != c_zb_error::NO_PROBLEM) ) {
		return;
	}
	if( p_code[0] != c_parser::OP_EQU ) {
		p_res->error_code = c_zb_error::ERROR_SYNTAX;
		return;
	}
	p_code++;
	p_res->error_code = expression.expression( result, p_code );
	if( p_res->error_code == c_zb_error::NO_PROBLEM ) {
		p_res->error_code = p_integer_array->set( num_elements, elements, result.get()->get_integer() );
	}
}

/* --------------------------------------------------------- */
void c_interpreter::process_rarray( const unsigned char* &p_code ) {
	c_variable_real_array *p_real_array;
	int elements[ MAX_ARRAY_DIMENTION ], num_elements;
	c_expression expression( p_res );
	c_value result;

	p_code++;
	p_real_array = reinterpret_cast<c_variable_real_array*>( c_parser::get_ptr( p_code ) );
	p_code += sizeof(void*);
	num_elements = dimension_check( p_code, elements );
	if( (num_elements == 0) || (p_res->error_code != c_zb_error::NO_PROBLEM) ) {
		return;
	}
	if( p_code[0] != c_parser::OP_EQU ) {
		p_res->error_code = c_zb_error::ERROR_SYNTAX;
		return;
	}
	p_code++;
	p_res->error_code = expression.expression( result, p_code );
	if( p_res->error_code == c_zb_error::NO_PROBLEM ) {
		p_res->error_code = p_real_array->set( num_elements, elements, result.get()->get_real() );
	}
}

/* --------------------------------------------------------- */
void c_interpreter::process_sarray( const unsigned char* &p_code ) {
	c_variable_string_array *p_string_array;
	int elements[ MAX_ARRAY_DIMENTION ], num_elements;
	c_expression expression( p_res );
	c_value result;
	const char *p_string_value;
	int length;

	p_code++;
	p_string_array = reinterpret_cast<c_variable_string_array*>( c_parser::get_ptr( p_code ) );
	p_code += sizeof(void*);
	num_elements = dimension_check( p_code, elements );
	if( (num_elements == 0) || (p_res->error_code != c_zb_error::NO_PROBLEM) ) {
		return;
	}
	if( p_code[0] != c_parser::OP_EQU ) {
		p_res->error_code = c_zb_error::ERROR_SYNTAX;
		return;
	}
	p_code++;
	p_res->error_code = expression.expression( result, p_code );
	if( p_res->error_code == c_zb_error::NO_PROBLEM ) {
		result.get()->get_string( p_string_value, length );
		p_res->error_code = p_string_array->set( num_elements, elements, p_string_value, length );
	}
}

/* --------------------------------------------------------- */
int c_interpreter::run( c_parser *p_line ) {
	int id;
	const unsigned char *p_code = p_line->get_precompiled_code();

	dlog.printf( "Start interpreter.\n" );
	for_info.clear();
	list.goto_bottom();

	p_res->f_program_running = true;

	while( p_res->f_program_running ) {
		switch( p_code[0] ) {
		case c_parser::PC_COMMAND:
			p_code++;
			memcpy( &id, p_code, sizeof(id) );
			p_code += sizeof(id);
			(command.get_entry(id))( p_code, p_res );
			break;
		case c_parser::PC_IVAR:
			process_ivar( p_code );
			break;
		case c_parser::PC_RVAR:
			process_rvar( p_code );
			break;
		case c_parser::PC_SVAR:
			process_svar( p_code );
			break;
		case c_parser::PC_IARRAY:
			process_iarray( p_code );
			break;
		case c_parser::PC_RARRAY:
			process_rarray( p_code );
			break;
		case c_parser::PC_SARRAY:
			process_sarray( p_code );
			break;
		case c_parser::PC_LABEL:
			dlog.printf( "  label\n" );
			p_code += 1 + sizeof(void*);
			break;
		case c_parser::PC_GOTO:
			dlog.printf( "  goto\n" );
			cmd_goto( p_code );
			break;
		case c_parser::PC_IF:
			dlog.printf( "  if\n" );
			cmd_if( p_code );
			break;
		case c_parser::PC_FOR:
			dlog.printf( "  for\n" );
			cmd_for( p_code );
			break;
		case c_parser::PC_NEXT:
			dlog.printf( "  next\n" );
			cmd_next( p_code );
			break;
		case c_parser::PC_GOSUB:
			dlog.printf( "  gosub\n" );
			cmd_gosub( p_code );
			break;
		case c_parser::PC_RETURN:
			dlog.printf( "  return\n" );
			cmd_return( p_code );
			break;
		case c_parser::PC_ENDMARK:
			list.goto_next_line();
			if( list.is_bottom() ) {
				p_res->f_program_running = false;
				break;
			}
			p_code = list.get_line();
			break;
		case c_parser::OP_COLON:
			p_code++;
			break;
		default:
			p_res->error_code = c_zb_error::ERROR_SYNTAX;
			break;
		}
		if( (p_code == 0) || (p_res->error_code != c_zb_error::NO_PROBLEM) ) {
			p_res->f_program_running = false;
		}
	}

	p_line->variable_release();
	for_info.clear();
	gosub_info.clear();
	p_res->p_callback->update_screen();
	dlog.printf( "End interpreter.\n" );
	return list.get_current_line_no();
}

/* --------------------------------------------------------- */
void c_interpreter::file_open( int descriptor, const char* p_file_name, const char *p_direction ) {

	if( (descriptor < 0) || (descriptor >= max_file_descriptor) ) {
		p_res->error_code = c_zb_error::ERROR_ILLEGAL_PARAMETER;
		return;
	}
	if( p_file[descriptor] ) {
		p_res->error_code = c_zb_error::ERROR_ALREADY_OPENED;
		return;
	}
	p_file[descriptor] = fopen( p_file_name, p_direction );
	if( p_file[descriptor] == 0 ) {
		p_res->error_code = c_zb_error::ERROR_DEVICE_IO_ERROR;
		return;
	}
}

/* --------------------------------------------------------- */
void c_interpreter::file_close( int descriptor ) {
	if( p_file[descriptor] ) {
		fclose( p_file[descriptor] );
		p_file[descriptor] = 0;
	}
}

/* --------------------------------------------------------- */
void c_interpreter::file_print( int descriptor, const char *p_string, int length ) {

	if( (descriptor < 0) || (descriptor >= max_file_descriptor) ) {
		p_res->error_code = c_zb_error::ERROR_ILLEGAL_PARAMETER;
		return;
	}
	if( p_file[descriptor] == 0 ) {
		p_res->error_code = c_zb_error::ERROR_INVALID_DESCRIPTOR;
		return;
	}
	if( fwrite( p_string, 1, length, p_file[descriptor] ) != static_cast<size_t>( length ) ) {
		p_res->error_code = c_zb_error::ERROR_DEVICE_IO_ERROR;
		return;
	}
}

/* --------------------------------------------------------- */
int c_interpreter::file_input( int descriptor, char *p_string, int size ) {

	if( (descriptor < 0) || (descriptor >= max_file_descriptor) ) {
		p_res->error_code = c_zb_error::ERROR_ILLEGAL_PARAMETER;
		return 0;
	}
	if( p_file[descriptor] == 0 ) {
		p_res->error_code = c_zb_error::ERROR_INVALID_DESCRIPTOR;
		return 0;
	}
	if( fgets( p_string, size, p_file[descriptor] ) == NULL ) {
		p_res->error_code = c_zb_error::ERROR_DEVICE_IO_ERROR;
		return 0;
	}
	return static_cast<int>( strlen( p_string ) );
}

/* --------------------------------------------------------- */
int c_interpreter::file_eof( int descriptor ) {

	if( (descriptor < 0) || (descriptor >= max_file_descriptor) ) {
		p_res->error_code = c_zb_error::ERROR_ILLEGAL_PARAMETER;
		return 0;
	}
	if( p_file[descriptor] == 0 ) {
		p_res->error_code = c_zb_error::ERROR_INVALID_DESCRIPTOR;
		return 0;
	}
	return( feof( p_file[descriptor] ) ? -1 : 0 );
}
