/* --------------------------------------------------------- */
/*	BASIC-E													 */
/*	Copyright(C)2006 t.hara									 */
/* ========================================================= */
/*	variable manager										 */
/* --------------------------------------------------------- */

#include "variable_manager.h"
#include "parser.h"
#include "zb_error.h"

/* --------------------------------------------------------- */
c_variable *c_variable_manager::search( c_variable *p_list, const char *p_name ) {

	while( p_list ) {
		if( strcmp( p_list->get_name(), p_name ) == 0 ) {
			break;
		}
		p_list = p_list->get_next();
	}
	return p_list;
}

/* --------------------------------------------------------- */
void c_variable_manager::line_release( const unsigned char *p_precompiled_code, int length ) {
	int i;
	c_variable *p;

	for( i = 0; i < length; i = c_parser::get_next_position( p_precompiled_code, i ) ) {
		if( c_parser::is_variable( p_precompiled_code[i] ) ) {
			p = reinterpret_cast<c_variable*>( c_parser::get_ptr( &p_precompiled_code[i+1] ) );
			p->release_ref();
		}
	}
}

/* --------------------------------------------------------- */
void c_variable_manager::array_clear( void ) {
	c_variable *p_list;

	for( p_list = array_list.get_next(); p_list != 0; p_list = p_list->get_next() ) {
		p_list->clear();
	}
}

/* --------------------------------------------------------- */
void c_variable_integer_array::clear( void ) {

	delete[] p_value;
	p_value = 0;
	dimension = 0;
}

/* --------------------------------------------------------- */
void c_variable_real_array::clear( void ) {

	delete[] p_value;
	p_value = 0;
	dimension = 0;
}

/* --------------------------------------------------------- */
void c_variable_string_array::clear( void ) {

	delete[] p_value;
	p_value = 0;
	dimension = 0;
}

/* --------------------------------------------------------- */
int c_variable_integer_array::get( int num_elements, int *p_element, int &v ) {
	int i, index;

	if( p_value == 0 ) {
		/* automatic DIM */
		int element[ MAX_ARRAY_DIMENTION ];
		for( i = 0; i < num_elements; i++ ) {
			element[i] = 10;
		}
		dim( num_elements, element );
	}
	if( dimension != num_elements ) return c_zb_error::ERROR_SUBSCRIPT_OUT_OF_RANGE;
	for( i = 0, index = 0; i < num_elements; i++ ) {
		if( static_cast<unsigned>( p_element[i] ) >= static_cast<unsigned>( counts[i] ) ) {
			return c_zb_error::ERROR_SUBSCRIPT_OUT_OF_RANGE;
		}
		index = (index * counts[i]) + p_element[i];
	}
	v = p_value[ index ];
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_variable_real_array::get( int num_elements, int *p_element, double &v ) {
	int i, index;

	if( p_value == 0 ) {
		/* automatic DIM */
		int element[ MAX_ARRAY_DIMENTION ];
		for( i = 0; i < num_elements; i++ ) {
			element[i] = 10;
		}
		dim( num_elements, element );
	}
	if( dimension != num_elements ) return c_zb_error::ERROR_SUBSCRIPT_OUT_OF_RANGE;
	for( i = 0, index = 0; i < num_elements; i++ ) {
		if( static_cast<unsigned>( p_element[i] ) >= static_cast<unsigned>( counts[i] ) ) {
			return c_zb_error::ERROR_SUBSCRIPT_OUT_OF_RANGE;
		}
		index = (index * counts[i]) + p_element[i];
	}
	v = p_value[ index ];
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_variable_string_array::get( int num_elements, int *p_element, const char* &p_string, int &length ) {
	int i, index;

	if( p_value == 0 ) {
		/* automatic DIM */
		int element[ MAX_ARRAY_DIMENTION ];
		for( i = 0; i < num_elements; i++ ) {
			element[i] = 10;
		}
		dim( num_elements, element );
	}
	if( dimension != num_elements ) return c_zb_error::ERROR_SUBSCRIPT_OUT_OF_RANGE;
	for( i = 0, index = 0; i < num_elements; i++ ) {
		if( static_cast<unsigned>( p_element[i] ) >= static_cast<unsigned>( counts[i] ) ) {
			return c_zb_error::ERROR_SUBSCRIPT_OUT_OF_RANGE;
		}
		index = (index * counts[i]) + p_element[i];
	}
	p_string = p_value[ index ].value;
	length = p_value[ index ].length;
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_variable_integer_array::set( int num_elements, int *p_element, int v ) {
	int i, index;

	if( p_value == 0 ) {
		/* automatic DIM */
		int element[ MAX_ARRAY_DIMENTION ];
		for( i = 0; i < num_elements; i++ ) {
			element[i] = 10;
		}
		dim( num_elements, element );
	}
	if( dimension != num_elements ) return c_zb_error::ERROR_SUBSCRIPT_OUT_OF_RANGE;
	for( i = 0, index = 0; i < num_elements; i++ ) {
		if( static_cast<unsigned>( p_element[i] ) >= static_cast<unsigned>( counts[i] ) ) {
			return c_zb_error::ERROR_SUBSCRIPT_OUT_OF_RANGE;
		}
		index = (index * counts[i]) + p_element[i];
	}
	p_value[ index ] = v;
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_variable_integer_array::dim( int num_elements, int *p_element ) {
	int i, index;

	if( p_value != 0 ) return c_zb_error::ERROR_REDIMENSIONED_ARRAY;
	if( num_elements > MAX_ARRAY_DIMENTION ) return c_zb_error::ERROR_SUBSCRIPT_OUT_OF_RANGE;
	for( i = 0, index = 1; i < num_elements; i++ ) {
		if( static_cast<unsigned>( p_element[i] ) >= static_cast<unsigned>( MAX_ELEMENT_MEMBER ) ) {
			return c_zb_error::ERROR_SUBSCRIPT_OUT_OF_RANGE;
		}
		counts[i] = p_element[i] + 1;
		index = index * counts[i];
	}
	p_value = new int[ index ];
	if( p_value == 0 ) {
		return c_zb_error::ERROR_OUT_OF_MEMORY;
	}
	memset( p_value, 0, sizeof(int) * index );
	dimension = num_elements;
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_variable_real_array::set( int num_elements, int *p_element, double v ) {
	int i, index;

	if( p_value == 0 ) {
		/* automatic DIM */
		int element[ MAX_ARRAY_DIMENTION ];
		for( i = 0; i < num_elements; i++ ) {
			element[i] = 10;
		}
		dim( num_elements, element );
	}
	if( dimension != num_elements ) return c_zb_error::ERROR_SUBSCRIPT_OUT_OF_RANGE;
	for( i = 0, index = 0; i < num_elements; i++ ) {
		if( static_cast<unsigned>( p_element[i] ) >= static_cast<unsigned>( counts[i] ) ) {
			return c_zb_error::ERROR_SUBSCRIPT_OUT_OF_RANGE;
		}
		index = (index * counts[i]) + p_element[i];
	}
	p_value[ index ] = v;
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_variable_real_array::dim( int num_elements, int *p_element ) {
	int i, index;

	if( p_value != 0 ) return c_zb_error::ERROR_REDIMENSIONED_ARRAY;
	if( num_elements > MAX_ARRAY_DIMENTION ) return c_zb_error::ERROR_SUBSCRIPT_OUT_OF_RANGE;
	for( i = 0, index = 1; i < num_elements; i++ ) {
		if( static_cast<unsigned>( p_element[i] ) >= static_cast<unsigned>( MAX_ELEMENT_MEMBER ) ) {
			return c_zb_error::ERROR_SUBSCRIPT_OUT_OF_RANGE;
		}
		counts[i] = p_element[i] + 1;
		index = index * counts[i];
	}
	p_value = new double[ index ];
	if( p_value == 0 ) {
		return c_zb_error::ERROR_OUT_OF_MEMORY;
	}
	memset( p_value, 0, sizeof(double) * index );
	dimension = num_elements;
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_variable_string_array::set( int num_elements, int *p_element, const char *p_string, int length ) {
	int i, index;

	if( p_value == 0 ) {
		/* automatic DIM */
		int element[ MAX_ARRAY_DIMENTION ];
		for( i = 0; i < num_elements; i++ ) {
			element[i] = 10;
		}
		dim( num_elements, element );
	}
	if( dimension != num_elements ) return c_zb_error::ERROR_SUBSCRIPT_OUT_OF_RANGE;
	for( i = 0, index = 0; i < num_elements; i++ ) {
		if( static_cast<unsigned>( p_element[i] ) >= static_cast<unsigned>( counts[i] ) ) {
			return c_zb_error::ERROR_SUBSCRIPT_OUT_OF_RANGE;
		}
		index = (index * counts[i]) + p_element[i];
	}
	p_value[ index ].length = length;
	if( length ) memcpy( p_value[ index ].value, p_string, length );
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_variable_string_array::dim( int num_elements, int *p_element ) {
	int i, index;

	if( p_value != 0 ) return c_zb_error::ERROR_REDIMENSIONED_ARRAY;
	if( num_elements > MAX_ARRAY_DIMENTION ) return c_zb_error::ERROR_SUBSCRIPT_OUT_OF_RANGE;
	for( i = 0, index = 1; i < num_elements; i++ ) {
		if( static_cast<unsigned>( p_element[i] ) >= static_cast<unsigned>( MAX_ELEMENT_MEMBER ) ) {
			return c_zb_error::ERROR_SUBSCRIPT_OUT_OF_RANGE;
		}
		counts[i] = p_element[i] + 1;
		index = index * counts[i];
	}
	p_value = new c_string_element[ index ];
	if( p_value == 0 ) {
		return c_zb_error::ERROR_OUT_OF_MEMORY;
	}
	dimension = num_elements;
	return c_zb_error::NO_PROBLEM;
}
